#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#


APPLICATION_NAME = 'openvpn3-indicator'
APPLICATION_TITLE = 'OpenVPN3 Indicator'
APPLICATION_ID = 'net.openvpn.openvpn3_indicator'
APPLICATION_SYSTEM_TAG = 'openvpn3-indicator'
APPLICATION_AUTHORS = [
        'Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>',
    ]
APPLICATION_URL = 'https://github.com/OpenVPN/openvpn3-indicator'
APPLICATION_VERSION = '0.1'
MANAGER_VERSION_MINIMUM = 20
MANAGER_VERSION_RECOMMENDED = 21
APPLICATION_DESCRIPTION_SHORT = 'Simple indicator application for OpenVPN3'
APPLICATION_DESCRIPTION_LONG = '''
This is a simple indicator application that controls OpenVPN3 tunnels.
It is based on D-Bus interface provided by OpenVPN3 Linux client.
It is a temporary solution until Network Manager supports OpenVPN3.
'''.strip()
CONFIGURATION_MIME_TYPE = 'application/x-openvpn-profile'



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, GObject, Gtk, Gio

from openvpn3_indicator.about import APPLICATION_NAME, CONFIGURATION_MIME_TYPE

DEFAULT_CONFIG_NAME = 'NEW'

def construct_configuration_select_dialog(name=None, on_import=None, on_cancel=None):
    dialog = Gtk.FileChooserDialog(action=Gtk.FileChooserAction.OPEN)
    dialog.set_position(Gtk.WindowPosition.CENTER)
    dialog.set_keep_above(True)
    dialog.set_icon_name(APPLICATION_NAME)
    dialog.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
            'Import', Gtk.ResponseType.OK,
        )
    ovpn_filter = Gtk.FileFilter()
    ovpn_filter.set_name('OpenVPN Configuration Files')
    ovpn_filter.add_mime_type(CONFIGURATION_MIME_TYPE)
    dialog.add_filter(ovpn_filter)
    text_filter = Gtk.FileFilter()
    text_filter.set_name('Text Files')
    text_filter.add_mime_type('text/plain')
    dialog.add_filter(text_filter)
    all_filter = Gtk.FileFilter()
    all_filter.set_name('All Files')
    all_filter.add_pattern('*')
    dialog.add_filter(all_filter)
    dialog.set_show_hidden(False)
    dialog.set_filter(ovpn_filter)

    def on_dialog_destroy(_object):
        if on_cancel is not None:
            on_cancel()

    def on_dialog_response(_object, response):
        if response == Gtk.ResponseType.OK:
            dialog.disconnect_by_func(on_dialog_destroy)
            configuration_path = dialog.get_filename()
            import_dialog = construct_configuration_import_dialog(name=name, path=configuration_path, on_import=on_import, on_cancel=on_cancel)
            import_dialog.set_visible(True)
        dialog.destroy()

    dialog.connect('destroy', on_dialog_destroy)
    dialog.connect('response', on_dialog_response)
    dialog.show_all()
    return dialog

def guess_configuration_name(path):
    #TODO: Try to guess the name from path contents
    pass

def construct_configuration_import_dialog(path, name=None, on_import=None, on_cancel=None):
    name = name or guess_configuration_name(path) or DEFAULT_CONFIG_NAME
    dialog = Gtk.Dialog('OpenVPN Configuration Import')
    dialog.set_position(Gtk.WindowPosition.CENTER)
    dialog.set_keep_above(True)
    dialog.set_icon_name(APPLICATION_NAME)
    dialog.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
            'Import', Gtk.ResponseType.OK,
        )
    content_area = dialog.get_content_area()
    content_area.set_property('hexpand', True)
    content_area.set_property('vexpand', True)
    grid = Gtk.Grid(row_spacing=10, column_spacing=10, vexpand=True, hexpand=True, margin_top=20, margin_right=20, margin_bottom=20, margin_left=20)

    grid.attach(Gtk.Label(label='Configuration File', hexpand=True, xalign=0, margin_right=10), 0, 0, 1, 1)
    grid.attach(Gtk.Label(label=path, hexpand=True), 1, 0, 1, 1)
    
    grid.attach(Gtk.Label(label='Configuration Name', hexpand=True, xalign=0, margin_right=10), 0, 1, 1, 1)
    entry = Gtk.Entry(hexpand=True)
    entry.set_text(name)
    entry.set_activates_default(True)
    grid.attach(entry, 1, 1, 1, 1)

    content_area.add(grid)

    def on_dialog_destroy(_object):
        if on_cancel is not None:
            on_cancel()

    def on_dialog_response(_object, response):
        if response == Gtk.ResponseType.OK:
            dialog.disconnect_by_func(on_dialog_destroy)
            if on_import is not None:
                name = entry.get_text()
                #TODO: Warn the user if config name is already used.
                #TODO: Shall we allow to have two configs with the same name?
                #TODO: Is it possible to overwrite previous config and keep id?
                #TODO: Or is it better to remove the old one, and create new with the same name?
                on_import(name=name, path=path)
        dialog.destroy()

    dialog.connect('destroy', on_dialog_destroy)
    dialog.connect('response', on_dialog_response)
    default = dialog.get_widget_for_response(response_id=Gtk.ResponseType.OK)
    default.set_can_default(True)
    default.grab_default()
    dialog.show_all()
    return dialog


def construct_configuration_remove_dialog(name, on_remove=None, on_cancel=None):
    dialog = Gtk.Dialog('OpenVPN Configuration Remove')
    dialog.set_position(Gtk.WindowPosition.CENTER)
    dialog.set_keep_above(True)
    dialog.set_icon_name(APPLICATION_NAME)
    dialog.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
            'Remove', Gtk.ResponseType.OK,
        )
    content_area = dialog.get_content_area()
    content_area.set_property('hexpand', True)
    content_area.set_property('vexpand', True)
    grid = Gtk.Grid(row_spacing=10, column_spacing=10, vexpand=True, hexpand=True, margin_top=20, margin_right=20, margin_bottom=20, margin_left=20)

    grid.attach(Gtk.Label(label=f'You are about to remove configuration {name}', hexpand=True, xalign=0, margin_right=10), 0, 0, 1, 1)

    content_area.add(grid)

    def on_dialog_destroy(_object):
        if on_cancel is not None:
            on_cancel()

    def on_dialog_response(_object, response):
        if response == Gtk.ResponseType.OK:
            dialog.disconnect_by_func(on_dialog_destroy)
            if on_remove is not None:
                on_remove()
        dialog.destroy()

    dialog.connect('destroy', on_dialog_destroy)
    dialog.connect('response', on_dialog_response)
    default = dialog.get_widget_for_response(response_id=Gtk.ResponseType.OK)
    default.set_can_default(True)
    default.grab_default()
    dialog.show_all()
    return dialog



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple GTK indicator GUI for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import argparse
import os
import pathlib
import re
import subprocess

parser = argparse.ArgumentParser()
parser.add_argument('--scriptdir', default=os.getcwd())
parser.add_argument('--outdir', default=os.getcwd())
args = parser.parse_args()
scriptdir=pathlib.Path(args.scriptdir)
outdir=pathlib.Path(args.outdir)
gitdir=scriptdir.parent

DISTROS = ['focal','jammy','noble','plucky','questing','resolute']


os.environ['TZ'] = 'UTC'
os.environ['LC_ALL'] = 'en_US.UTF-8'

subprocess.run(['git', 'config', '--global', '--add', 'safe.directory', str(gitdir)])

timestamp_run = subprocess.run(['git', '-C', str(gitdir), 'log', '-n', '1', '--format=format:%cd', '--date=format-local:%Y%m%d%H%M%S'], capture_output=True)
TIMESTAMP = str(timestamp_run.stdout, 'utf-8').strip()

icon_path = gitdir / 'share' / 'icons'
ICON_THEMES = [ str(p.relative_to(icon_path)) for p in icon_path.glob('*') if p.is_dir() ]
ICONS = [ str(p.relative_to(icon_path.parent)) for p in icon_path.glob('**/*.svg') if p.is_file() ]

mime_path = gitdir / 'share' / 'mime'
MIMES = [ str(p.relative_to(mime_path.parent)) for p in mime_path.glob('**/*.xml') if p.is_file() ]

app_path = gitdir / 'share' / 'applications'
APPS = [ str(p.relative_to(app_path.parent)) for p in app_path.glob('**/*.desktop') if p.is_file() ]

man_path = gitdir / 'share' / 'man'
MANS = [ str(p.relative_to(man_path.parent)) for p in man_path.glob('**/*.[1-8]') if p.is_file() ]


for DISTRO in DISTROS:

    NAME      = 'openvpn3-indicator'
    MAINTAINER = 'Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>'
    PRIORITY  = 'optional'
    SECTION   = 'net'
    VERSION   = f'0.1.{TIMESTAMP}'
    RELEASE   = DISTRO
    SUMMARY   = 'Simple GTK indicator GUI for OpenVPN3'
    DESCRIPTION = '''This project adds a simple GTK indicator application that can be used to control OpenVPN3 tunnels.
It is based on D-Bus interface provided by OpenVPN3 Linux client and does not require elevated privileges to use.
It should be considered as a temporary work-around until Network Manager implements support for OpenVPN 3, or OpenVPN provides a graphical interface for Linux users.
'''
    LICENSE   = 'AGPL-3.0'
    URL       = 'https://github.com/OpenVPN/openvpn3-indicator'
    COMPAT = '10'
    BUILDREQUIRES = [
            f'debhelper (>= {COMPAT})',
            'desktop-file-utils',
            'make',
            'python3',
            'sed',
            'zip',
        ]
    REQUIRES  = [
            'openvpn3 | openvpn3-client',
            'python3-dbus',
            'python3-gi',
            'python3-secretstorage',
            'python3-setproctitle',
            'gir1.2-ayatanaappindicator3-0.1',
        ]

    SOURCES = [
        'src',
        'scripts',
        'Makefile',
        'LICENSE',
        'README.md',
        'share',
    ]
    SOURCECODE = outdir / f'openvpn3-indicator_{VERSION}.orig.tar.gz'
    SOURCECODE.parent.mkdir(parents=True, exist_ok=True)
    source_run = subprocess.run(['tar', '--create', '--auto-compress', '--file', SOURCECODE, '--transform', f'flags=r;s|^|openvpn3-indicator-{VERSION}/|', '--directory', str(gitdir) ] + SOURCES)

    NEWLINE='\n'

    changelog_run = subprocess.run(['git', '-C', str(gitdir), 'log', '-n', '1', '--format=format:'+NAME+' ('+VERSION+'-'+re.sub(r'%','%%',RELEASE)+') '+DISTRO+'; urgency=medium%n%n  * %s%b%n%n -- %an <%ae>  %cd', '--date=format-local:%a, %d %b %Y %H:%M:%S +0000'], capture_output=True)
    CHANGELOG = str(changelog_run.stdout, 'utf-8').strip()

    pkgdir = outdir / f'openvpn3-indicator-{VERSION}'

    debdir = pkgdir / 'debian'

    debdir.mkdir(parents=True, exist_ok=True)

    (debdir / 'control').write_text(
f'''Source: {NAME}
Maintainer: {MAINTAINER}
Priority: {PRIORITY}
Section: {SECTION}
Build-Depends: {", ".join(sorted(BUILDREQUIRES))}
Standards-Version: 4.0.0
Homepage: {URL}

Package: {NAME}
Architecture: all
Depends: {", ".join(sorted(REQUIRES))}
Description: 
{NEWLINE.join(["  "+line for line in DESCRIPTION.split(NEWLINE)])}
''')
    (debdir / 'source').mkdir(parents=True, exist_ok=True)
    (debdir / 'source' / 'format').write_text(
f'''3.0 (quilt)
''')
    (debdir / 'rules').write_text(re.sub(r'    ','\t',
f'''#!/usr/bin/make -f
clean:

build:
    make DESTDIR=debian/{NAME} BINDIR=/usr/bin DATADIR=/usr/share HARDCODE_PYTHON=/usr/bin/python3 all

binary:
    make DESTDIR=debian/{NAME} BINDIR=/usr/bin DATADIR=/usr/share HARDCODE_PYTHON=/usr/bin/python3 package
    dh_gencontrol
    dh_builddeb
'''))
    (debdir / 'changelog').write_text(
f'''{CHANGELOG}
''')
    (debdir / 'compat').write_text(
f'''{COMPAT}
''')
    (debdir / 'copyright').write_text( (gitdir / 'LICENSE').read_text() )

    subprocess.run(['dpkg-buildpackage', '-us', '-uc', '-ui', '-S'], cwd=pkgdir)



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import logging
import time
import uuid

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gio

from openvpn3_indicator.about import *

###
#
# MultiNotifier
#
###

class MultiNotifier():
    @property
    def identifier(self):
        return self._identifier
    def sub_identifier(self, identifier):
        return f'{self.identifier}-{identifier}'
    @property
    def  application(self):
        return self._application

    def __init__(self, application, identifier):
        self._identifier = identifier
        self._application = application
        self._notifiers = dict()
        self._pending = list()
        self.default_icon = f'{APPLICATION_NAME}'
        self.default_title = f'{APPLICATION_TITLE}'
        self.default_body = None
        self.default_category = None
        self.default_priority = Gio.NotificationPriority.NORMAL
        self.default_timespan = None
        self.invalid = False

    def invalidate(self):
        self.invalid = True

    class Notifier():
        @property
        def parent(self):
            return self._parent
        @property
        def identifier(self):
            return self._identifier
        @property
        def sent(self):
            return self._sent
        @property
        def timeout(self):
            return self._timeout

        def __init__(self, parent, identifier, active=False, icon=None, title=None, body=None, category=None, priority=None, timespan=None, mute_repetitions=False):
            self._parent = parent
            self._identifier = identifier
            self._active = active
            self._icon = icon or self.parent.default_icon
            self._title = title or self.parent.default_title
            self._body = body or self.parent.default_body
            self._category = category or self.parent.default_category
            self._priority = priority or self.parent.default_priority
            self._timespan = timespan or self.parent.default_timespan
            self._mute_repetitions = mute_repetitions
            self._last_sent = None
            self._sent = None
            self._timeout = None

        def close(self):
            if self.parent:
                self.parent.del_notifier(self)
        @property
        def active(self):
            return self._active
        @active.setter
        def active(self, active):
            active = bool(active)
            if self._active != active:
                self._active = active
                if not self._active:
                    if self._sent and self.parent:
                        self.parent.application.withdraw_notification(self.identifier)
                    self._sent = None
                    self._timeout = None
                if self.parent:
                    self.parent.invalidate()
        @property
        def icon(self):
            return self._icon
        @icon.setter
        def icon(self, icon):
            icon = str(icon)
            if self._icon != icon:
                self._icon = icon
                if self.parent and self.active:
                    self.parent.invalidate()
        @property
        def title(self):
            return self._title
        @title.setter
        def title(self, title):
            title = str(title)
            if self._title != title:
                self._title = title
                if self.parent and self.active:
                    self.parent.invalidate()
        @property
        def body(self):
            return self._body
        @body.setter
        def body(self, body):
            body = str(body)
            if self._body != body:
                self._body = body
                if self.parent and self.active:
                    self.parent.invalidate()
        @property
        def category(self):
            return self._category
        @category.setter
        def category(self, category):
            category = str(category)
            if self._category != category:
                self._category = category
                if self.parent and self.active:
                    self.parent.invalidate()
        @property
        def priority(self):
            return self._priority
        @priority.setter
        def priority(self, priority):
            if self._priority != priority:
                self._priority = priority
                if self.parent and self.active:
                    self.parent.invalidate()
        @property
        def timespan(self):
            return self._timespan
        @timespan.setter
        def timespan(self, timespan):
            if self._timespan != timespan:
                self._timespan = timespan
                if self.parent and self.active:
                    self.parent.invalidate()

    def new_notifier(self, identifier=None, **kwargs):
        identifier = self.sub_identifier(identifier or str(uuid.uuid4()))
        notifier = self.Notifier(self, identifier=identifier, **kwargs)
        self._notifiers[identifier] = notifier
        logging.debug(f'Created Notifier {identifier}')
        if notifier.active:
            self.invalidate()
        return notifier
    def del_notifier(self, notifier):
        if notifier.parent == self:
            if notifier.identifier in self._notifiers:
                if notifier.active:
                    self._pending.append(notifier)
                del self._notifiers[notifier.identifier]
                logging.debug(f'Destroyed Notifier {notifier.identifier}')
            notifier._parent = None

    def commit_notifier(self, notifier):
        if notifier.timeout is not None and notifier.timeout < time.monotonic():
            notifier.active = False
        if notifier.active and notifier.sent is None:
            send_description = (notifier.icon, notifier.title, notifier.body, notifier.category, notifier.priority)
            if notifier._mute_repetitions and notifier._last_sent == send_description:
                notifier.active = False
            else:
                notifier._sent = time.monotonic()
                notifier._timeout = notifier.timespan and notifier.sent + notifier.timespan
                notifier._last_sent = send_description
                target = Gio.Notification.new(notifier.title or '')
                if notifier.icon is not None:
                    icon = Gio.Icon.new_for_string(notifier.icon)
                    target.set_icon(icon)
                if notifier.title is not None:
                    target.set_title(notifier.title)
                if notifier.body is not None:
                    target.set_body(notifier.body)
                if notifier.category is not None:
                    target.set_category(notifier.category)
                if notifier.priority is not None:
                    target.set_priority(notifier.priority)
                self.application.send_notification(notifier.identifier, target)
        if not notifier.active and notifier.sent is not None:
            notifier._sent = None
            notifier._timeout = None
            self.application.withdraw_notification(notifier.identifier)

    def update(self):
        for notifier in self._notifiers.values():
            self.commit_notifier(notifier)
        new_pending = list()
        for notifier in self._pending:
            self.commit_notifier(notifier)
            if notifier.active:
                new_pending.append(notifier)
        self._pending = new_pending
        self.invalid = False

    def close(self):
        for notifier in list(self._notifiers.values()):
            notifier.close()



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple GTK indicator GUI for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

if __name__ == '__main__':
    import openvpn3_indicator
    openvpn3_indicator.main()



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, Gtk

from openvpn3_indicator.about import APPLICATION_NAME

def construct_notification_dialog(parent=None, title=None, message=None, message_type=Gtk.MessageType.INFO):
    dialog = Gtk.MessageDialog(
        parent=parent,
        flags=Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
        type=message_type,
        buttons=Gtk.ButtonsType.OK
    )

    if title:
        dialog.set_title(title)
    else:
        type_names = {
            Gtk.MessageType.INFO: "Info",
            Gtk.MessageType.WARNING: "Warning", 
            Gtk.MessageType.ERROR: "Error",
            Gtk.MessageType.QUESTION: "Question"
        }
        type_name = type_names.get(message_type, "Notification")
        dialog.set_title(f"{APPLICATION_NAME} - {type_name}")

    if message:
        dialog.set_markup(GLib.markup_escape_text(message))

    return dialog


def show_notification_dialog(parent=None, title=None, message=None, message_type=Gtk.MessageType.INFO):
    dialog = construct_notification_dialog(parent, title, message, message_type)
    response = dialog.run()
    dialog.destroy()
    return response


def show_info_notification(parent=None, title=None, message=None):
    return show_notification_dialog(parent, title, message, Gtk.MessageType.INFO)


def show_warning_notification(parent=None, title=None, message=None):
    return show_notification_dialog(parent, title, message, Gtk.MessageType.WARNING)


def show_error_notification(parent=None, title=None, message=None):
    return show_notification_dialog(parent, title, message, Gtk.MessageType.ERROR)


def show_question_notification(parent=None, title=None, message=None):
    dialog = Gtk.MessageDialog(
        parent=parent,
        flags=Gtk.DialogFlags.MODAL | Gtk.DialogFlags.DESTROY_WITH_PARENT,
        type=Gtk.MessageType.QUESTION,
        buttons=Gtk.ButtonsType.YES_NO
    )

    if title:
        dialog.set_title(title)
    else:
        dialog.set_title(f"{APPLICATION_NAME} - Question")

    if message:
        dialog.set_markup(GLib.markup_escape_text(message))

    response = dialog.run()
    dialog.destroy()
    return response


def construct_error_dialog(parent=None, title=None, message=None):
    return construct_notification_dialog(parent, title, message, Gtk.MessageType.ERROR)


def show_error_dialog(parent=None, title=None, message=None):
    return show_error_notification(parent, title, message)


#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import collections

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, GObject, Gtk, Gio

from openvpn3_indicator.about import APPLICATION_NAME


CredentialsUserInput = collections.namedtuple(
        'CredentialsUserInput',
        ['name', 'mask', 'value', 'can_store']
    )


def construct_credentials_dialog(name, user_inputs, allow_store=True, on_connect=None, on_cancel=None, remain_active=None):
    dialog = Gtk.Dialog('OpenVPN Credentials')
    dialog.set_position(Gtk.WindowPosition.CENTER)
    dialog.set_keep_above(True)
    dialog.set_icon_name(APPLICATION_NAME)
    dialog.add_buttons(
            Gtk.STOCK_CONNECT, Gtk.ResponseType.OK,
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL,
        )
    content_area = dialog.get_content_area()
    content_area.set_property('hexpand', True)
    content_area.set_property('vexpand', True)
    grid = Gtk.Grid(row_spacing=10, column_spacing=10, vexpand=True, hexpand=True, margin_top=20, margin_right=20, margin_bottom=20, margin_left=20)
    introduction = f'Session {name} requires providing credentials'
    grid.attach(Gtk.Label(label=introduction, hexpand=True, margin_bottom=10), 0, 0, 2, 1)
    row = 1
    entries = list()
    default_store = False
    can_store = False
    for user_input in user_inputs:
        label = Gtk.Label(label=user_input.name, hexpand=True, xalign=0, margin_right=10)
        grid.attach(label, 0, row, 1, 1)
        entry = Gtk.Entry(hexpand=True)
        if user_input.mask:
            entry.set_visibility(False)
        if user_input.can_store:
            can_store = True
        if user_input.value is not None:
            entry.set_text(user_input.value)
            default_store = True
        entry.set_activates_default(True)
        grid.attach(entry, 1, row, 1, 1)
        entries.append(entry)
        row += 1
    if can_store and allow_store:
        store_button = Gtk.CheckButton(label='Store credentials', hexpand=True, margin_top=10)
        store_button.set_active(default_store)
        grid.attach(store_button, 0, row, 2, 1)
    content_area.add(grid)

    def on_dialog_destroy(_object):
        if on_cancel is not None:
            on_cancel()

    def on_dialog_response(_object, response):
        if response == Gtk.ResponseType.OK:
            dialog.disconnect_by_func(on_dialog_destroy)
            if on_connect is not None:
                results = list()
                for user_input, entry in zip(user_inputs, entries):
                    results.append(CredentialsUserInput(name=user_input.name, mask=user_input.mask, value=entry.get_text(), can_store=user_input.can_store))
                store = None
                if can_store and allow_store:
                    store = store_button.get_active()
                on_connect(user_inputs=results, store=store)
        dialog.destroy()

    def on_schedule():
        if remain_active():
            GLib.timeout_add(1000, on_schedule)
        else:
            try:
                dialog.disconnect_by_func(on_dialog_destroy)
                dialog.destroy()
            except:
                pass

    dialog.connect('destroy', on_dialog_destroy)
    dialog.connect('response', on_dialog_response)
    default = dialog.get_widget_for_response(response_id=Gtk.ResponseType.OK)
    default.set_can_default(True)
    default.grab_default()
    dialog.show_all()
    if remain_active is not None:
        GLib.timeout_add(1000, on_schedule)
    return dialog



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

import logging
import sys

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import GLib, Gtk

from openvpn3_indicator.about import APPLICATION_ID
from openvpn3_indicator.dialogs.configuration import construct_configuration_select_dialog

class Test(Gtk.Application):
    def __init__(self):
        Gtk.Application.__init__(self,
            application_id=APPLICATION_ID,
            )
        self.connect('startup', self.on_startup)
        self.connect('activate', self.on_activate)

    def on_activate(self, *args, **kwargs):
        pass
    def on_startup(self, *args, **kwargs):
        self.hold()

        dialog = construct_configuration_select_dialog(on_import=self.action_import, on_cancel=self.action_quit)
        dialog.set_visible(True)

        GLib.timeout_add(1000, self.on_schedule)
        GLib.timeout_add(60000, self.action_quit)

    def on_schedule(self, *args, **kwargs):
        GLib.timeout_add(1000, self.on_schedule)

    def action_import(self, *args, **kwargs):
        print(f'Result: {args} {kwargs}')
        self.action_quit()

    def action_quit(self, *args, **kwargs):
        self.release()

if __name__ == '__main__':
    logging.basicConfig(level = logging.DEBUG)
    test = Test()
    test.run(sys.argv)



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

###
#
# Main
#
###


def main(args=None):
    import logging
    import sys
    import traceback
    from openvpn3_indicator.about import APPLICATION_NAME
    from openvpn3_indicator.application import Application

    try:
        import setproctitle
        setproctitle.setproctitle(f'{APPLICATION_NAME}')
    except ImportError:
        logging.debug(traceback.format_exc())
        logging.error('Failed to import setproctitle module')

    try:
        import gi
        gi.require_version('Gtk', '3.0')
        from gi.repository import GLib, GObject, Gtk, Gio
    except:
        logging.critical('OpenVPN Indicator requires GTK to run. Please install required libraries.')
        sys.exit(1)

    from openvpn3_indicator.dialogs.system_checks import construct_dbus_missing_dialog, construct_appindicator_missing_dialog, construct_openvpn_missing_dialog, construct_secret_storage_missing_dialog
    
    try:
        import dbus
        from dbus.mainloop.glib import DBusGMainLoop
    except:
        logging.critical('OpenVPN Indicator requires D-Bus to run. Please install required libraries.')
        dialog = construct_dbus_missing_dialog()
        dialog.set_visible(True)
        dialog.run()
        sys.exit(1)

    try:
        try:
            gi.require_version('AyatanaAppIndicator3', '0.1')
            from gi.repository import AyatanaAppIndicator3 as AppIndicator3
        except:
            gi.require_version('AppIndicator3', '0.1')
            from gi.repository import AppIndicator3
    except:
        logging.critical('OpenVPN Indicator requires AppIndicator to run. Please install AppIndicator plugin for your desktop.')
        dialog = construct_appindicator_missing_dialog()
        dialog.set_visible(True)
        dialog.run()
        sys.exit(1)

    try:
        import openvpn3
    except:
        logging.critical('OpenVPN Indicator requires OpenVPN3 python library to run. Please install OpenVPN3.')
        dialog = construct_openvpn_missing_dialog()
        dialog.set_visible(True)
        dialog.run()
        sys.exit(1)

    try:
        import secretstorage
    except:
        logging.critical('OpenVPN Indicator requires Secret Storage to run. Please install required libraries.')
        dialog = construct_secret_storage_missing_dialog()
        dialog.set_visible(True)
        dialog.run()
        sys.exit(1)

    application = Application()
    if args is None:
        args = sys.argv
    application.run(args)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, GObject, Gtk, Gio

from openvpn3_indicator.about import APPLICATION_NAME, APPLICATION_TITLE, APPLICATION_VERSION, APPLICATION_AUTHORS, APPLICATION_URL, APPLICATION_DESCRIPTION_SHORT


def construct_appindicator_missing_dialog():
    dialog = Gtk.MessageDialog()
    dialog.set_position(Gtk.WindowPosition.CENTER)
    dialog.set_keep_above(True)
    dialog.set_icon_name(APPLICATION_NAME)
    dialog.add_buttons(
            Gtk.STOCK_OK, Gtk.ResponseType.OK,
        )
    dialog.set_markup('<b>AppIndicator not available</b>')
    dialog.format_secondary_text('OpenVPN Indicator requires AppIndicator to run. Please install AppIndicator plugin for your desktop.')

    def on_dialog_response(_object, response):
        dialog.destroy()

    dialog.connect('response', on_dialog_response)
    dialog.show_all()
    return dialog


def construct_dbus_missing_dialog():
    dialog = Gtk.MessageDialog()
    dialog.set_position(Gtk.WindowPosition.CENTER)
    dialog.set_keep_above(True)
    dialog.set_icon_name(APPLICATION_NAME)
    dialog.add_buttons(
            Gtk.STOCK_OK, Gtk.ResponseType.OK,
        )
    dialog.set_markup('<b>D-Bus not available</b>')
    dialog.format_secondary_text('OpenVPN Indicator requires D-Bus to run. Please install required libraries.')

    def on_dialog_response(_object, response):
        dialog.destroy()

    dialog.connect('response', on_dialog_response)
    dialog.show_all()
    return dialog


def construct_openvpn_missing_dialog():
    dialog = Gtk.MessageDialog()
    dialog.set_position(Gtk.WindowPosition.CENTER)
    dialog.set_keep_above(True)
    dialog.set_icon_name(APPLICATION_NAME)
    dialog.add_buttons(
            Gtk.STOCK_OK, Gtk.ResponseType.OK,
        )
    dialog.set_markup('<b>OpenVPN3 not available</b>')
    dialog.format_secondary_text('OpenVPN Indicator requires OpenVPN3 python library to run. Please install OpenVPN3.')

    def on_dialog_response(_object, response):
        dialog.destroy()

    dialog.connect('response', on_dialog_response)
    dialog.show_all()
    return dialog


def construct_secret_storage_missing_dialog():
    dialog = Gtk.MessageDialog()
    dialog.set_position(Gtk.WindowPosition.CENTER)
    dialog.set_keep_above(True)
    dialog.set_icon_name(APPLICATION_NAME)
    dialog.add_buttons(
            Gtk.STOCK_OK, Gtk.ResponseType.OK,
        )
    dialog.set_markup('<b>Secret Storage not available</b>')
    dialog.format_secondary_text('OpenVPN Indicator requires Secret Storage to run. Please install required libraries.')

    def on_dialog_response(_object, response):
        dialog.destroy()

    dialog.connect('response', on_dialog_response)
    dialog.show_all()
    return dialog



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import logging
import uuid

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
try:
    gi.require_version('AyatanaAppIndicator3', '0.1')
    from gi.repository import AyatanaAppIndicator3 as AppIndicator3
except (ValueError, ImportError):
    gi.require_version('AppIndicator3', '0.1')
    from gi.repository import AppIndicator3

from openvpn3_indicator.about import *

###
#
# MultiIndicator
#
###

class MultiIndicator():

    @property
    def identifier(self):
        return self._identifier

    def sub_identifier(self, num):
        if num == 0:
            return f'{self.identifier}'
        return f'{self.identifier}-{num}'

    def sub_indicator(self, num):
        while len(self._sub_indicators) <= num:
            sub = AppIndicator3.Indicator.new(
                self.sub_identifier(len(self._sub_indicators)),
                self.default_icon,
                self.default_category
                )
            sub.set_ordering_index(num)
            self._sub_indicators.append(sub)
        return self._sub_indicators[num]

    def __init__(self, identifier):
        self._identifier = identifier
        self._sub_indicators = list()
        self._indicators = dict()
        self.default_icon = f'{APPLICATION_NAME}'
        self.default_description = f'{APPLICATION_TITLE}'
        self.default_title = f'{APPLICATION_TITLE}'
        self.default_category = AppIndicator3.IndicatorCategory.SYSTEM_SERVICES
        self.invalid = False

    def invalidate(self):
        self.invalid = True

    class Indicator():

        @property
        def parent(self):
            return self._parent

        @property
        def identifier(self):
            return self._identifier

        def __init__(self, parent, identifier, active=False, icon=None, description=None, title=None, order_key=None, menu=None):
            self._parent = parent
            self._identifier = identifier
            self._active = active
            self._icon = icon or self.parent.default_icon
            self._description = description or self.parent.default_description
            self._title = title or self.parent.default_title
            self._order_key = order_key or self.identifier
            self._menu = menu

        def close(self):
            if self.parent:
                self.parent.del_indicator(self)

        @property
        def active(self):
            return self._active
        @active.setter
        def active(self, active):
            active = bool(active)
            if self._active != active:
                self._active = active
                if self.parent:
                    self.parent.invalidate()

        @property
        def icon(self):
            return self._icon
        @icon.setter
        def icon(self, icon):
            icon = str(icon)
            if self._icon != icon:
                self._icon = icon
                if self.parent and self.active:
                    self.parent.invalidate()

        @property
        def description(self):
            return self._description
        @description.setter
        def description(self, description):
            description = str(description)
            if self._description != description:
                self._description = description
                if self.parent and self.active:
                    self.parent.invalidate()

        @property
        def title(self):
            return self._title
        @title.setter
        def title(self, title):
            title = str(title)
            if self._title != title:
                self._title = title
                if self.parent and self.active:
                    self.parent.invalidate()

        @property
        def order_key(self):
            return self._order_key
        @order_key.setter
        def order_key(self, order_key):
            order_key = str(order_key)
            if self._order_key != order_key:
                self._order_key = order_key
                if self.parent and self.active:
                    self.parent.invalidate()

        @property
        def menu(self):
            return self._menu
        @menu.setter
        def menu(self, menu):
            if self._menu != menu:
                self._menu = menu
                if self.parent and self.active:
                    self.parent.invalidate()

    def new_indicator(self, **kwargs):
        identifier = str(uuid.uuid4())
        indicator = self.Indicator(self, identifier, **kwargs)
        self._indicators[identifier] = indicator
        logging.debug(f'Created Indicator {identifier}')
        if indicator.active:
            self.invalidate()
        return indicator

    def del_indicator(self, indicator):
        if indicator.parent == self:
            if indicator.identifier in self._indicators:
                del self._indicators[indicator.identifier]
                if indicator.active:
                    self.invalidate()
                logging.debug(f'Destroyed Indicator {indicator.identifier}')
            indicator._parent = None

    def commit_indicator(self, indicator, num):
        target = self.sub_indicator(num)
        target.set_icon_full(indicator.icon, indicator.description)
        target.set_title(indicator.title)
        if indicator.menu:
            target.set_menu(indicator.menu)
        else:
            target.set_menu(Gtk.Menu())
        target.set_status(AppIndicator3.IndicatorStatus.ACTIVE)

    def hide_indicator(self, num):
        target = self.sub_indicator(num)
        target.set_menu(Gtk.Menu())
        target.set_status(AppIndicator3.IndicatorStatus.PASSIVE)

    def reset(self):
        for indicator in self._sub_indicators:
            indicator.set_menu(Gtk.Menu())
            indicator.set_status(AppIndicator3.IndicatorStatus.PASSIVE)
        self._sub_indicators = list()
        self.invalid = True

    def update(self):
        if self.invalid:
            logging.debug('Repairing Indicators')
            indicators = list()
            for indicator in self._indicators.values():
                if indicator.active:
                    indicators.append(indicator)
            num = 0
            for indicator in sorted(indicators, key=lambda i : i.order_key):
                self.commit_indicator(indicator, num)
                num += 1
            for other in range(num, len(self._sub_indicators)):
                self.hide_indicator(other)
            self.invalid=False

    def close(self):
        for indicator in list(self._indicators.values()):
            indicator.close()



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import collections

from openvpn3.constants import StatusMinor

from openvpn3_indicator.about import APPLICATION_NAME

DEFAULT_ICON = f'{APPLICATION_NAME}-idle'
DEFAULT_DESCRIPTION = 'Unknown'

StatusDescription = collections.namedtuple(
        'StatusDescription',
        ['minor', 'name', 'icon', 'description']
    )

status_descriptions = dict()

def register_status(name, icon, description):
    if hasattr(StatusMinor, name):
        minor = getattr(StatusMinor, name)
        desc = StatusDescription(minor=minor, name=name, icon=icon, description=description)
        status_descriptions[minor] = desc

register_status('UNSET',                  'idle',       'Unknown')
register_status('CFG_ERROR',              'idle-error', 'Configuration error')
register_status('CFG_OK',                 'loading',    'Connecting')
register_status('CFG_INLINE_MISSING',     'idle-error', 'Configuration error')
register_status('CFG_REQUIRE_USER',       'loading',    'Authentication required')
register_status('CONN_INIT',              'loading',    'Connecting')
register_status('CONN_CONNECTING',        'loading',    'Connecting')
register_status('CONN_CONNECTED',         'active',     'Connected')
register_status('CONN_DISCONNECTING',     'loading',    'Disconnecting')
register_status('CONN_DISCONNECTED',      'idle',       'Disconnected')
register_status('CONN_FAILED',            'idle-error', 'Connection failed')
register_status('CONN_AUTH_FAILED',       'idle-error', 'Authentication failed')
register_status('CONN_RECONNECTING',      'loading',    'Connecting')
register_status('CONN_PAUSING',           'loading',    'Pausing')
register_status('CONN_PAUSED',            'paused',     'Paused')
register_status('CONN_RESUMING',          'loading',    'Connecting')
register_status('CONN_DONE',              'idle',       'Disconnected')
register_status('SESS_NEW',               'loading',    'Connecting')
register_status('SESS_BACKEND_COMPLETED', 'loading',    'Connecting')
register_status('SESS_REMOVED',           'idle',       'Disconnected')
register_status('SESS_AUTH_USERPASS',     'loading',    'Authentication required')
register_status('SESS_AUTH_CHALLENGE',    'loading',    'Authentication required')
register_status('SESS_AUTH_URL',          'loading',    'Authentication required')
register_status('PKCS11_SIGN',            'loading',    'Authentication required')
register_status('PKCS11_ENCRYPT',         'loading',    'Authentication required')
register_status('PKCS11_DECRYPT',         'loading',    'Authentication required')
register_status('PKCS11_VERIFY',          'loading',    'Authentication required')
register_status('PROC_STARTED',           'loading',    'Process started')
register_status('PROC_STOPPED',           'idle-error', 'Process stopped')
register_status('PROC_KILLED',            'idle-error', 'Process killed')


def get_status_icon(major, minor):
    if minor in status_descriptions:
        icon = status_descriptions[minor].icon
        if icon:
            return f'{APPLICATION_NAME}-{icon}'
    return DEFAULT_ICON


def get_status_description(major, minor):
    if minor in status_descriptions:
        description = status_descriptions[minor].description
        if description:
            return description
    return DEFAULT_DESCRIPTION



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple GTK indicator GUI for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import argparse
import os
import pathlib
import re
import subprocess

parser = argparse.ArgumentParser()
parser.add_argument('--scriptdir', default=os.getcwd())
parser.add_argument('--outdir', default=os.getcwd())
args = parser.parse_args()
scriptdir=pathlib.Path(args.scriptdir)
outdir=pathlib.Path(args.outdir)
gitdir=scriptdir.parent



os.environ['TZ'] = 'UTC'
os.environ['LC_ALL'] = 'en_US.UTF-8'

subprocess.run(['git', 'config', '--global', '--add', 'safe.directory', str(gitdir)])

timestamp_run = subprocess.run(['git', '-C', str(gitdir), 'log', '-n', '1', '--format=format:%cd', '--date=format-local:%Y%m%d%H%M%S'], capture_output=True)
TIMESTAMP = str(timestamp_run.stdout, 'utf-8').strip()

icon_path = gitdir / 'share' / 'icons'
ICON_THEMES = [ str(p.relative_to(icon_path)) for p in icon_path.glob('*') if p.is_dir() ]
ICONS = [ str(p.relative_to(icon_path.parent)) for p in icon_path.glob('**/*.svg') if p.is_file() ]

mime_path = gitdir / 'share' / 'mime'
MIMES = [ str(p.relative_to(mime_path.parent)) for p in mime_path.glob('**/*.xml') if p.is_file() ]

schema_path = gitdir / 'share' / 'glib-2.0' / 'schemas'
SCHEMAS = [ str(p.relative_to(schema_path.parent.parent)) for p in schema_path.glob('**/*.gschema.xml') if p.is_file() ]

app_path = gitdir / 'share' / 'applications'
APPS = [ str(p.relative_to(app_path.parent)) for p in app_path.glob('**/*.desktop') if p.is_file() ]

man_path = gitdir / 'share' / 'man'
MANS = [ str(p.relative_to(man_path.parent)) for p in man_path.glob('**/*.[1-8]') if p.is_file() ]

NAME      = 'openvpn3-indicator'
VERSION   = f'0.1.{TIMESTAMP}'
RELEASE   = '1' #'1%{?dist}'
SUMMARY   = 'Simple GTK indicator GUI for OpenVPN3'
DESCRIPTION = '''
This project adds a simple GTK indicator application that can be used to control OpenVPN3 tunnels.
It is based on D-Bus interface provided by OpenVPN3 Linux client and does not require elevated privileges to use.
It should be considered as a temporary work-around until Network Manager implements support for OpenVPN 3, or OpenVPN provides a graphical interface for Linux users.
'''
LICENSE   = 'AGPL-3.0'
URL       = 'https://github.com/OpenVPN/openvpn3-indicator'
BUILDARCH = 'noarch'
BUILDREQUIRES = ' '.join(sorted([
        'desktop-file-utils',
        'make',
        'python3',
        'sed',
        'zip',
    ]))
REQUIRES  = ' '.join(sorted([
        'openvpn3-client',
        'python3-dbus',
        'python3-secretstorage',
        'python3-setproctitle',
        'gnome-shell-extension-appindicator',
    ]))

SOURCES = [
    'src',
    'scripts',
    'Makefile',
    'LICENSE',
    'README.md',
    'share',
]
SOURCECODE = outdir / f'openvpn3-indicator-{VERSION}.tar.gz'
source_run = subprocess.run(['tar', '--create', '--auto-compress', '--file', SOURCECODE, '--transform', f'flags=r;s|^|openvpn3-indicator-{VERSION}/|', '--directory', str(gitdir) ] + SOURCES)

PREP = '\n'.join([
        '%setup'
    ])

BUILD = '\n'.join([
        'make DESTDIR=%{buildroot} BINDIR=%{_bindir} DATADIR=%{_datadir} HARDCODE_PYTHON=/usr/bin/python3 all',
    ])

INSTALL = '\n'.join([
        'make DESTDIR=%{buildroot} BINDIR=%{_bindir} DATADIR=%{_datadir} package',
    ] + [
        'desktop-file-validate %{buildroot}%{_datadir}/'+path for path in APPS
    ])

POST = '\n'.join([
        'touch --no-create %{_datadir}/icons/'+theme+' || :' for theme in ICON_THEMES
    ] + [
        'xdg-mime install --mode system %{_datadir}/'+mime for mime in MIMES
    ])

POSTTRANS = '\n'.join([
        'update-desktop-database %{_datadir}/applications || :',
        'update-mime-database %{_datadir}/mime || :',
        'glib-compile-schemas %{_datadir}/glib-2.0/schemas || :',
    ] + [
        'gtk-update-icon-cache --silent %{_datadir}/icons/'+theme+' || :' for theme in ICON_THEMES
    ])

PREUN = '\n'.join([
        'xdg-mime uninstall --mode system %{_datadir}/'+mime for mime in MIMES
    ])

POSTUN = '\n'.join([
        'update-desktop-database %{_datadir}/applications || :',
        'update-mime-database %{_datadir}/mime || :',
        'glib-compile-schemas %{_datadir}/glib-2.0/schemas || :',
    ] + [
        'touch --no-create %{_datadir}/icons/'+theme+' || :' for theme in ICON_THEMES
    ] + [
        'gtk-update-icon-cache --silent %{_datadir}/icons/'+theme+' || :' for theme in ICON_THEMES
    ])

FILES = '\n'.join([
        '%license LICENSE',
        '%doc README.md',
        '%{_bindir}/%{name}',
        '/etc/xdg/autostart/%{name}.desktop'
    ] + [
        '%{_datadir}/'+path for path in APPS + ICONS + MIMES + SCHEMAS
    ] + [
        '%{_datadir}/'+path+'.gz' for path in MANS
    ])


changelog_run = subprocess.run(['git', '-C', str(gitdir), 'log', '-n', '1', '--format=format:* %cd %an <%ae> - '+VERSION+'-'+re.sub(r'%','%%',RELEASE)+'%n- %s%b%n', '--date=format-local:%a %b %d %Y'], capture_output=True)
CHANGELOG = str(changelog_run.stdout, 'utf-8').strip()

(outdir / 'openvpn3-indicator.spec').write_text(f'''
Name: {NAME}
Version: {VERSION}
Release: {RELEASE}
Summary: {SUMMARY}
License: {LICENSE}
URL: {URL}
BuildArch: {BUILDARCH}
BuildRequires: {BUILDREQUIRES}
Requires: {REQUIRES}
Source: {SOURCECODE}

%description

{DESCRIPTION}

%prep

{PREP}

%build

{BUILD}

%install

{INSTALL}

%post

{POST}

%posttrans

{POSTTRANS}

%preun

{PREUN}

%postun

{POSTUN}

%files

{FILES}

%changelog

{CHANGELOG}

''')



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

import logging
import sys

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import GLib, Gtk

from openvpn3_indicator.about import APPLICATION_ID
from openvpn3_indicator.dialogs.notification import (
    show_info_notification,
    show_warning_notification,
    show_error_notification,
    show_question_notification,
    construct_notification_dialog
)

class Test(Gtk.Application):
    def __init__(self):
        Gtk.Application.__init__(self,
            application_id=APPLICATION_ID,
            )
        self.connect('startup', self.on_startup)
        self.connect('activate', self.on_activate)
        self.test_step = 0

    def on_activate(self, *args, **kwargs):
        pass

    def on_startup(self, *args, **kwargs):
        self.hold()
        self.run_next_test()

    def run_next_test(self):
        if self.test_step == 0:
            print("Testing INFO notification...")
            show_info_notification(
                title="Test Info",
                message="This is an info notification test"
            )
        elif self.test_step == 1:
            print("Testing WARNING notification...")
            show_warning_notification(
                title="Test Warning",
                message="This is a warning notification test"
            )
        elif self.test_step == 2:
            print("Testing ERROR notification...")
            show_error_notification(
                title="Test Error",
                message="This is an error notification test"
            )
        elif self.test_step == 3:
            print("Testing QUESTION notification...")
            response = show_question_notification(
                title="Test Question",
                message="This is a question notification test. Do you want to continue?"
            )
            print(f"Question response: {response}")
        elif self.test_step == 4:
            print("Testing construct_notification_dialog...")
            dialog = construct_notification_dialog(
                title="Test Custom Dialog",
                message="This is a custom notification dialog test",
                message_type=Gtk.MessageType.INFO
            )
            dialog.connect('response', self.on_dialog_response)
            dialog.show()
            return
        else:
            print("All tests completed!")
            self.action_quit()
            return

        self.test_step += 1
        GLib.timeout_add(500, self.run_next_test)

    def on_dialog_response(self, dialog, response_id):
        print(f"Custom dialog response: {response_id}")
        dialog.destroy()
        self.test_step += 1
        GLib.timeout_add(500, self.run_next_test)

    def action_quit(self, *args, **kwargs):
        print("Test completed successfully!")
        self.release()

if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    print("Starting notification dialog tests...")
    print("Each dialog will appear sequentially. Click OK to proceed to the next test.")
    test = Test()
    test.run(sys.argv)



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

import logging
import sys

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import GLib, Gtk

from openvpn3_indicator.about import *
from openvpn3_indicator.multi_indicator import MultiIndicator

class Test(Gtk.Application):
    def __init__(self):
        Gtk.Application.__init__(self,
            application_id=APPLICATION_ID,
            )
        self.connect('startup', self.on_startup)
        self.connect('activate', self.on_activate)

    def on_activate(self, *args, **kwargs):
        pass
    def on_startup(self, *args, **kwargs):
        self.hold()
        self.multi_indicator = MultiIndicator('Test App')

        self.first_indicator = self.multi_indicator.new_indicator()
        self.first_indicator.icon = 'openvpn3-indicator-idle'
        self.first_indicator.description = 'Test Description'
        self.first_indicator.title = 'Test Title'
        self.first_indicator.order_key = '1'
        menu = Gtk.Menu()
        menu_item = Gtk.MenuItem.new_with_label('Quit')
        menu_item.connect('activate', self.action_quit)
        menu.append(menu_item)
        menu.show_all()
        self.first_indicator.menu = menu
        self.first_indicator.active = True

        self.second_indicator = self.multi_indicator.new_indicator()
        self.second_indicator.icon = 'openvpn3-indicator-active'
        self.second_indicator.description = 'Test Description'
        self.second_indicator.title = 'Test Title'
        self.second_indicator.order_key = '2'
        self.second_indicator.menu = menu
        self.second_indicator.active=True
        
        GLib.timeout_add(1000, self.on_schedule)
        GLib.timeout_add(5000, self.action_quit)

    def on_schedule(self, *args, **kwargs):
        self.multi_indicator.update()
        GLib.timeout_add(1000, self.on_schedule)

    def action_quit(self, *args, **kwargs):
        self.release()

if __name__ == '__main__':
    logging.basicConfig(level = logging.DEBUG)
    test = Test()
    test.run(sys.argv)



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import logging
import secretstorage
import traceback

from openvpn3_indicator.about import APPLICATION_NAME, APPLICATION_TITLE

###
#
# CredentialStore
#
###


class CredentialStore:

    @property
    def application_name(self):
        return self._application_name

    @property
    def application_description(self):
        return self._application_description

    @property
    def secret_collection(self):
        if self._secret_collection is None:
            try:
                secret_connection = secretstorage.dbus_init()
                self._secret_collection = secretstorage.get_default_collection(secret_connection)
            except:  # TODO: Catch only expected exceptions
                logging.debug(traceback.format_exc())
                logging.error('Failed to connect with Secret Storage')
        return self._secret_collection

    @property
    def secret_collection_unlocked(self):
        collection = self.secret_collection
        if collection is not None:
            if collection.is_locked():
                try:
                    collection.unlock()
                    logging.info('Unlocked Secret Storage')
                except:  # TODO: Catch only expected exceptions
                    logging.debug(traceback.format_exc())
                    logging.error('Failed to unlock Secret Storage')
                    collection = None
        return collection

    def attrs(self):
        attrs = dict()
        attrs['application'] = self.application_name
        return attrs

    def __init__(self, application_name=APPLICATION_NAME, application_description=APPLICATION_TITLE):
        self._application_name = application_name
        self._application_description = application_description
        self._secret_collection = None

    def __getitem__(self, key):
        return self.Credentials(self, key)

    def keys(self):
        collection = self.secret_collection_unlocked
        result = set()
        if collection:
            try:
                for item in collection.search_items(self.attrs()):
                    config = item.get_attributes().get('config', None)
                    if config is not None:
                        result.add(config)
            except:  # TODO: Catch only expected exceptions
                logging.debug(traceback.format_exc())
                logging.error('Failed to list Secret Storage')
        return sorted(list(result))

    class Credentials:

        @property
        def parent(self):
            return self._parent

        @property
        def config_id(self):
            return self._config_id

        def __init__(self, parent, config_id):
            self._parent = parent
            self._config_id = str(config_id)

        def attrs(self, key=None):
            attrs = self.parent.attrs()
            attrs['config'] = self.config_id
            if key is not None:
                attrs['key'] = str(key)
            return attrs

        def label(self, key):
            key = str(key)
            return f'{self.parent.application_description} {self.config_id} {key}'

        def __setitem__(self, key, item):
            key = str(key)
            item = str(item)
            collection = self.parent.secret_collection_unlocked
            if collection:
                try:
                    label = self.label(key)
                    if item is not None:
                        collection.create_item(
                            label,
                            self.attrs(key),
                            bytes(str(item), 'utf-8'),
                            replace=True)
                        logging.info(f'Stored secret {label} in Secret Storage')
                    else:
                        for item in collection.search_items(self.attrs(key)):
                            item.delete()
                        logging.info(f'Removed secret {label} from Secret Storage')
                except:  # TODO: Catch only expected exceptions
                    logging.debug(traceback.format_exc())
                    logging.error('Failed to write to Secret Storage')

        def __getitem__(self, key):
            key = str(key)
            collection = self.parent.secret_collection_unlocked
            if collection:
                try:
                    items = list(collection.search_items(self.attrs(key)))
                    if len(items) > 1:
                        logging.warning(f'There are multiple entries for {self.label(key)} in Secret Storage')
                    if len(items) > 0:
                        return str(items[0].get_secret(), 'utf-8')
                    logging.info(f'Retrieved secret {self.label(key)} from Secret Storage')
                except:  # TODO: Catch only expected exceptions
                    logging.debug(traceback.format_exc())
                    logging.error('Failed to read from Secret Storage')

        def __delitem__(self, key):
            key = str(key)
            collection = self.parent.secret_collection_unlocked
            if collection:
                try:
                    for item in collection.search_items(self.attrs(key)):
                        item.delete()
                    logging.info(f'Removed secret {self.label(key)} from Secret Storage')
                except:  # TODO: Catch only expected exceptions
                    logging.debug(traceback.format_exc())
                    logging.error('Failed to delete from Secret Storage')

        def keys(self):
            collection = self.parent.secret_collection_unlocked
            result = set()
            if collection:
                try:
                    for item in collection.search_items(self.attrs()):
                        key = item.get_attributes().get('key', None)
                        if key is not None:
                            result.add(key)
                except:  # TODO: Catch only expected exceptions
                    logging.debug(traceback.format_exc())
                    logging.error('Failed to list Secret Storage')
            return sorted(list(result))



#!/usr/bin/env python3
# vim:ts=4:sts=4:sw=4:expandtab

#
# openvpn3-indicator - Simple indicator application for OpenVPN3.
# Copyright (C) 2024 Grzegorz Gutowski <grzegorz.gutowski@uj.edu.pl>
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License,
# or any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.
# If not, see <https://www.gnu.org/licenses/>.
#

import gettext
import logging
import pathlib
import re
import sys
import time
import traceback
import webbrowser

import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib, GObject, Gtk, Gio

import dbus
from dbus.mainloop.glib import DBusGMainLoop
try:
    gi.require_version('AyatanaAppIndicator3', '0.1')
    from gi.repository import AyatanaAppIndicator3 as AppIndicator3
except:
    gi.require_version('AppIndicator3', '0.1')
    from gi.repository import AppIndicator3

import openvpn3

from openvpn3_indicator.about import APPLICATION_ID, APPLICATION_VERSION, APPLICATION_NAME, APPLICATION_TITLE, APPLICATION_SYSTEM_TAG
from openvpn3_indicator.about import MANAGER_VERSION_MINIMUM, MANAGER_VERSION_RECOMMENDED
from openvpn3_indicator.multi_indicator import MultiIndicator
from openvpn3_indicator.multi_notifier import MultiNotifier
from openvpn3_indicator.credential_store import CredentialStore
from openvpn3_indicator.dialogs.about import construct_about_dialog
from openvpn3_indicator.dialogs.system_checks import construct_appindicator_missing_dialog
from openvpn3_indicator.dialogs.credentials import CredentialsUserInput, construct_credentials_dialog
from openvpn3_indicator.dialogs.configuration import construct_configuration_select_dialog, construct_configuration_import_dialog, construct_configuration_remove_dialog
from openvpn3_indicator.dialogs.notification import show_error_dialog, show_warning_notification, show_info_notification
from openvpn3_indicator.status import get_status_icon, get_status_description


#TODO: Which input slots should not be stored ? (OTPs, etc.)
#TODO: Understand better the possible session state changes
#TODO: Collect and present session logs and stats
#TODO: Implement other than AppIndicator ways to have system tray icon
#TODO: /usr/share/metainfo ?
#TODO: Understand mimetype icons inheritance
#TODO: Prepare localization

DEFAULT_CONFIG_NAME = gettext.gettext('UNKNOWN')
DEFAULT_SESSION_NAME = gettext.gettext('UNKNOWN')

###
#
# Application
#
###

class Application(Gtk.Application):
    def __init__(self):
        Gtk.Application.__init__(self,
            application_id=APPLICATION_ID,
            flags=Gio.ApplicationFlags.HANDLES_OPEN,
            )
        self.settings = Gio.Settings.new(APPLICATION_ID)
        self.add_main_option('version', ord('V'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Show version and exit", None)
        self.add_main_option('verbose', ord('v'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Show more info", None)
        self.add_main_option('debug', ord('d'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Show debug info", None)
        self.add_main_option('silent', ord('s'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Show less info", None)
        self.clear_secret_storage = False
        self.add_main_option('clear-secret-storage', ord('c'), GLib.OptionFlags.NONE, GLib.OptionArg.NONE, "Remove all data stored in secret storage", None)
        self.connect('handle-local-options', self.on_handle_local_options)
        self.connect('startup', self.on_startup)
        self.connect('activate', self.on_activate)
        self.connect('open', self.on_open)
        self.connect('shutdown', self.on_shutdown)

    def on_handle_local_options(self, application, options):
        options = options.end().unpack()
        level=logging.WARNING
        if options.get('version', False):
            print(f'{APPLICATION_NAME} {APPLICATION_VERSION}')
            return 0
        if options.get('clear-secret-storage', False):
            self.clear_secret_storage = True
        if options.get('debug', False):
            level = logging.DEBUG
        elif options.get('silent', False):
            level = logging.INFO
        elif options.get('verbose', False):
            level = logging.ERROR
        logging.basicConfig(level = level)
        return -1

    def on_activate(self, data):
        self.info(f'Activate')

    def on_open(self, application, files, n_files, hint):
        self.info(f'Open {n_files} {hint}')
        for file in files:
            config_path = file.get_path()
            self.action_config_open(config_path)

    def on_shutdown(self, application):
        self.info('Shutdown')
        if hasattr(self, 'multi_indicator'):
            self.multi_indicator.close()
        if hasattr(self, 'multi_notifier'):
            self.multi_notifier.close()

    def on_startup(self, data):
        self.info(f'Startup')
        DBusGMainLoop(set_as_default=True)

        bus = dbus.Bus()
        self.session_bus = bus
        notifier_fail_count = 0
        while True:
            try:
                bus.get_name_owner('org.kde.StatusNotifierWatcher')
                break
            except dbus.exceptions.DBusException:
                notifier_fail_count += 1
                if notifier_fail_count > 20:
                    logging.critical('OpenVPN Indicator requires AppIndicator to run. Please install AppIndicator plugin for your desktop.')
                    dialog = construct_appindicator_missing_dialog()
                    dialog.set_visible(True)
                    dialog.run()
                    sys.exit(1)
                else:
                    time.sleep(0.5)

        self.multi_notifier = MultiNotifier(self, f'{APPLICATION_NAME}')
        self.notifiers = dict()
        self.session_bus.add_signal_receiver(
            self.on_status_notifier_watcher_owner_changed,
            signal_name='NameOwnerChanged',
            dbus_interface='org.freedesktop.DBus',
            bus_name='org.freedesktop.DBus',
            arg0='org.kde.StatusNotifierWatcher',
        )

        self.dbus = dbus.SystemBus()
        self.config_manager = openvpn3.ConfigurationManager(self.dbus)
        self.session_manager = openvpn3.SessionManager(self.dbus)
        self.session_manager.SessionManagerCallback(self.on_session_manager_event)

        # Detect if the config manager version is v21 or newer
        # TODO: This can be simplified once the openvpn3 module provides
        #       a version query API
        self.manager_version = 0
        cmgr_obj = self.dbus.get_object('net.openvpn.v3.configuration','/net/openvpn/v3/configuration')
        cmgr_prop = dbus.Interface(cmgr_obj, dbus_interface='org.freedesktop.DBus.Properties')
        cmgr_peer = dbus.Interface(cmgr_obj, dbus_interface='org.freedesktop.DBus.Peer')
        self.manager_version = 9999
        try:
            cmgr_peer.Ping()
            try:
                cmgr_version = str(cmgr_prop.Get('net.openvpn.v3.configuration','version'))
            except dbus.exceptions.DBusException:
                self.debug(f'Waiting for backend to start')
                time.sleep(0.5)
                cmgr_version = str(cmgr_prop.Get('net.openvpn.v3.configuration','version'))
            if cmgr_version.startswith('git:'):
                # development version: presume all features are available
                # and use a high version number
                pass
            elif cmgr_version.startswith('v'):
                # Version identifiers may cary a "release label",
                # like v19_beta, v22_dev
                self.manager_version = int(re.split(r'[^0-9]', cmgr_version[1:], 1)[0])
        except:
            self.debug(traceback.format_exc())
            self.warning(f'Backend version check failed')
        if self.manager_version < MANAGER_VERSION_MINIMUM:
            self.error(f'You are using version {self.manager_version} of OpenVPN3 software which is not supported. Consider an upgrade to a newer version. We recommend version {MANAGER_VERSION_RECOMMENDED}.', notify=True)
        elif self.manager_version < MANAGER_VERSION_RECOMMENDED:
            self.warning(f'You are using version {self.manager_version} of OpenVPN3 software. Consider an upgrade to a newer version. We recommend version {MANAGER_VERSION_RECOMMENDED}.', notify=True)
        self.debug(f'Running with manager version {self.manager_version}')

        self.credential_store = CredentialStore()
        if self.clear_secret_storage:
            for config in self.credential_store.keys():
                credentials = self.credential_store[config]
                for key in list(credentials.keys()):
                    self.info(f'Removing entry {key} from secret storage')
                    del credentials[key]

        self.configs = dict()
        self.sessions = dict()
        self.sessions_connected = set()
        self.config_names = dict()
        self.name_configs = dict()
        self.config_sessions = dict()
        self.session_configs = dict()
        self.failed_authentications = set()
        self.session_dialogs = dict()
        self.session_statuses = dict()

        self.multi_indicator = MultiIndicator(f'{APPLICATION_NAME}')
        self.default_indicator = self.multi_indicator.new_indicator()
        self.default_indicator.icon=f'{APPLICATION_NAME}-idle'
        self.default_indicator.description=f'{APPLICATION_TITLE}'
        self.default_indicator.title=f'{APPLICATION_TITLE}'
        self.default_indicator.order_key='0'
        self.default_indicator.active=True
        self.indicators = dict()

        self.last_invalid = time.monotonic()
        self.invalid_sessions = True
        self.invalid_ui = True

        self.startup_config_id = None
        self.startup_config_name = None
        try:
            startup_action = self.settings.get_string('startup-action')
            self.debug(f'Startup action: {startup_action}')
            if startup_action == 'RESTART':
                self.startup_config_id = self.settings.get_string('most-recent-configuration-id')
            start_id = re.match(r'STARTID:(?P<id>.*)', startup_action)
            if start_id:
                self.startup_config_id = start_id.group('id')
            start_name = re.match(r'STARTNAME:(?P<name>.*)', startup_action)
            if start_name:
                self.startup_config_name = start_name.group('name')
        except:
            pass
        if self.startup_config_id or self.startup_config_name:
            self.info(f'Startup configuration set to {self.startup_config_id or self.startup_config_name}')

        GLib.timeout_add(1000, self.on_schedule)
        self.hold()

    def on_status_notifier_watcher_owner_changed(self, name, old_owner, new_owner):
        old_owner = str(old_owner)
        new_owner = str(new_owner)
        self.info(f'StatusNotifierWatcher owner changed from {old_owner or "<none>"} to {new_owner or "<none>"}')
        if not new_owner:
            return
        self.multi_indicator.reset()
        self.invalid_ui = True

    def refresh_ui(self):
        if self.invalid_ui:
            new_indicators = dict()
            for session_id in self.sessions:
                indicator = self.indicators.get(session_id, None)
                if indicator is None:
                    session_name = self.get_session_name(session_id)
                    indicator = self.multi_indicator.new_indicator()
                    indicator.icon = self.session_icon(session_id)
                    indicator.description = f'{APPLICATION_TITLE}: {session_name}'
                    indicator.title = f'{APPLICATION_TITLE}: {session_name}'
                    indicator.order_key = f'1-{session_name}-{session_id}'
                    indicator.active = True
                new_indicators[session_id] = indicator
            new_notifiers = dict()
            for session_id in self.sessions:
                notifier = self.notifiers.get(session_id, None)
                if notifier is None:
                    session_name = self.get_session_name(session_id)
                    notifier = self.multi_notifier.new_notifier(f'session-{session_id}-status', mute_repetitions=True)
                    notifier.icon = self.session_icon(session_id)
                    notifier.title = f'{APPLICATION_TITLE}: {session_name}'
                    notifier.body = self.session_description(session_id)
                    notifier.active = False
                new_notifiers[session_id] = notifier
            for session_id, indicator in self.indicators.items():
                if session_id not in new_indicators:
                    indicator.close()
            for session_id, notifier in self.notifiers.items():
                if session_id not in new_notifiers:
                    notifier.close()
            if len(new_indicators) == 0:
                self.default_indicator.active = True
                #TODO: Change icon, description, etc. Based on what?
                self.default_indicator.menu = self.construct_idle_menu()
            else:
                self.default_indicator.active = False
            self.indicators = new_indicators
            for session_id, indicator in self.indicators.items():
                if session_id is not None:
                    #TODO: Change icon, description, etc. based on status
                    indicator.menu = self.construct_session_menu(session_id)
            self.multi_indicator.update()
            self.notifiers = new_notifiers
            self.invalid_ui = False

    def refresh_sessions(self):
        if self.invalid_sessions:
            new_session_ids = set()
            try:
                new_sessions = dict()
                for session in self.session_manager.FetchAvailableSessions():
                    session_id = str(session.GetPath())
                    if session_id not in self.sessions:
                        new_sessions[session_id] = session
                        session.StatusChangeCallback(lambda major, minor, message: self.on_session_event(session_id, major, minor, message))
                        new_session_ids.add(session_id)
                    else:
                        new_sessions[session_id] = self.sessions[session_id]
                new_configs = dict()
                for config in self.config_manager.FetchAvailableConfigs():
                    config_id = str(config.GetPath())
                    if config_id not in self.configs:
                        new_configs[config_id] = config
                    else:
                        new_configs[config_id] = self.configs[config_id]
                new_config_names = dict()
                for config_id, config in new_configs.items():
                    config_name = str(config.GetConfigName())
                    new_config_names[config_id] = config_name
                new_config_sessions = dict()
                new_session_configs = dict()
                for config_id, config_name in new_config_names.items():
                    new_config_sessions[config_id] = list()
                    for session_id in self.session_manager.LookupConfigName(config_name):
                        session_id = str(session_id)
                        new_config_sessions[config_id].append(session_id)
                        new_session_configs[session_id] = config_id
                new_session_statuses = dict()
                for session_id, session in new_sessions.items():
                    status = session.GetStatus()
                    new_session_statuses[session_id] = {
                        'major' : openvpn3.StatusMajor(status['major']),
                        'minor' : openvpn3.StatusMinor(status['minor']),
                        'message' : str(status['message']),
                    }
                self.sessions = new_sessions
                self.configs = new_configs
                self.config_names = new_config_names
                self.name_configs = dict([(value, key) for key,value in new_config_names.items()])
                self.config_sessions = new_config_sessions
                self.session_configs = new_session_configs
                self.session_statuses = new_session_statuses

                self.debug(f'Configs: {sorted(self.configs.keys())}')
                self.debug(f'Sessions: {sorted(self.sessions.keys())}')
                self.debug(f'Config names: {self.config_names}')
                self.debug(f'Config sessions: {self.config_sessions}')
                self.debug(f'Session configs: {self.session_configs}')
                self.debug(f'Session statuses: {self.session_statuses}')
                self.invalid_sessions = False
                self.invalid_ui = True
            except: #TODO: Catch only expected exceptions
                self.debug(traceback.format_exc())
                self.warning(f'Session list refresh failed')
            for session_id in new_session_ids:
                session_status = self.session_statuses[session_id]
                self.on_session_event(session_id, session_status['major'], session_status['minor'], session_status['message'])
            for session_id, dialog in list(self.session_dialogs.items()):
                if session_id not in self.sessions:
                    dialog.destroy()
                    if session_id not in self.sessions:
                        del self.session_dialogs[session_id]

    def get_config_name(self, config_id):
        return self.config_names.get(config_id, DEFAULT_CONFIG_NAME)

    def get_session_name(self, session_id):
        return self.get_config_name(self.session_configs.get(session_id, ''))

    def action_settings_startup(self, _object, value):
        self.settings.set_string('startup-action', value)
        self.invalid_ui = True
        self.refresh_ui()

    def construct_menu_settings_startup(self):
        startup_action = self.settings.get_string('startup-action') or ''
        menu = Gtk.Menu()
        menu_action = ''
        menu_title = gettext.gettext('No Connection')
        if startup_action == menu_action:
            menu_title += ' ✓'
        menu_item = Gtk.MenuItem.new_with_label(menu_title)
        menu_item.connect('activate', self.action_settings_startup, menu_action)
        menu.append(menu_item)
        menu_action = 'RESTART'
        menu_title = gettext.gettext('Restart Connection')
        if startup_action == menu_action:
            menu_title += ' ✓'
        menu_item = Gtk.MenuItem.new_with_label(menu_title)
        menu_item.connect('activate', self.action_settings_startup, menu_action)
        menu.append(menu_item)
        for config_name, config_id in sorted(self.name_configs.items()):
            menu_action = f'STARTNAME:{config_name}'
            menu_title = gettext.gettext('Start {name}').format(name=config_name)
            if startup_action == menu_action:
                menu_title += ' ✓'
            menu_item = Gtk.MenuItem.new_with_label(menu_title)
            menu_item.connect('activate', self.action_settings_startup, menu_action)
            menu.append(menu_item)
        return menu

    def construct_menu_config(self, config_id):
        menu = Gtk.Menu()
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Connect'))
        menu_item.connect('activate', self.action_config_connect, config_id)
        menu.append(menu_item)
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Remove'))
        menu_item.connect('activate', self.action_config_remove, config_id)
        menu.append(menu_item)
        return menu

    def construct_menu_session(self, session_id):
        menu = Gtk.Menu()
        status = self.session_statuses[session_id]
        major = status['major']
        minor = status['minor']
        menu_item = Gtk.MenuItem.new_with_label(self.get_session_name(session_id))
        menu.append(menu_item)

        if False: #TODO: When does it make sense to allow explicit Connect?
            menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Connect'))
            menu_item.connect('activate', self.action_session_connect, session_id)
            menu.append(menu_item)
        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CONN_CONNECTED == minor:
            menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Pause'))
            menu_item.connect('activate', self.action_session_pause, session_id)
            menu.append(menu_item)
        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CONN_PAUSED == minor:
            menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Resume'))
            menu_item.connect('activate', self.action_session_resume, session_id)
            menu.append(menu_item)
        if True:
            menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Restart'))
            menu_item.connect('activate', self.action_session_restart, session_id)
            menu.append(menu_item)
        if True:
            menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Disconnect'))
            menu_item.connect('activate', self.action_session_disconnect, session_id)
            menu.append(menu_item)
        return menu

    def construct_session_menu(self, session_id):
        menu = self.construct_menu_session(session_id)
        menu.append(Gtk.SeparatorMenuItem())
        add_separator = False
        for config_name, config_id in sorted(self.name_configs.items()):
            if len(self.config_sessions[config_id]) == 0:
                config_menu = self.construct_menu_config(config_id)
                menu_item = Gtk.MenuItem.new_with_label(config_name)
                menu_item.set_submenu(config_menu)
                menu.append(menu_item)
                add_separator = True
        if add_separator:
            menu.append(Gtk.SeparatorMenuItem())
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Import Config'))
        menu_item.connect('activate', self.action_config_import)
        menu.append(menu_item)
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Startup Settings'))
        menu_item.set_submenu(self.construct_menu_settings_startup())
        menu.append(menu_item)
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('About'))
        menu_item.connect('activate', self.action_about)
        menu.append(menu_item)
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Quit'))
        menu_item.connect('activate', self.action_quit)
        menu.append(menu_item)
        menu.show_all()
        return menu

    def construct_idle_menu(self):
        menu = Gtk.Menu()
        for config_name, config_id in sorted(self.name_configs.items()):
            session_ids = self.config_sessions[config_id]
            if len(session_ids) > 0:
                for session_id in session_ids:
                    #TODO: Add some information on session status to menu items (perhaps in the title?)
                    session = self.sessions[session_id]
                    session_menu = self.construct_menu_session(session_id)
                    menu_item = Gtk.MenuItem.new_with_label(config_name)
                    menu_item.set_submenu(session_menu)
                    menu.append(menu_item)
            else:
                config_menu = self.construct_menu_config(config_id)
                menu_item = Gtk.MenuItem.new_with_label(config_name)
                menu_item.set_submenu(config_menu)
                menu.append(menu_item)
        if len(self.name_configs) > 0:
            menu.append(Gtk.SeparatorMenuItem())
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Import Config'))
        menu_item.connect('activate', self.action_config_import)
        menu.append(menu_item)
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Startup Settings'))
        menu_item.set_submenu(self.construct_menu_settings_startup())
        menu.append(menu_item)
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('About'))
        menu_item.connect('activate', self.action_about)
        menu.append(menu_item)
        menu_item = Gtk.MenuItem.new_with_label(gettext.gettext('Quit'))
        menu_item.connect('activate', self.action_quit)
        menu.append(menu_item)
        menu.show_all()
        return menu

    def session_icon(self, session_id):
        status = self.session_statuses[session_id]
        major = status['major']
        minor = status['minor']
        return get_status_icon(major, minor)

    def session_description(self, session_id):
        status = self.session_statuses[session_id]
        major = status['major']
        minor = status['minor']
        return get_status_description(major, minor)

    def notify_session_change(self, session_id):
        indicator = self.indicators.get(session_id, None)
        if indicator:
            indicator.icon = self.session_icon(session_id)
        notifier = self.notifiers.get(session_id, None)
        if notifier:
            notifier.active = False
            notifier.icon = self.session_icon(session_id)
            notifier.body = self.session_description(session_id)
            notifier.timespan = 3
            notifier.active = True

    def on_session_manager_event(self, event):
        self.info(f'Session Manager Event {event}')
        event_type = event.GetType()
        if openvpn3.SessionManagerEventType.SESS_CREATED == event_type:
            self.invalid_sessions = True
        elif openvpn3.SessionManagerEventType.SESS_DESTROYED == event_type:
            self.invalid_sessions = True

    def on_network_manager_event(self, event):
        self.info(f'Network Manager Event {event}')

    def can_store_input_slot(self, input_slot):
        type, group = input_slot.GetTypeGroup()
        result = type == openvpn3.ClientAttentionType.CREDENTIALS and group in [
                openvpn3.ClientAttentionGroup.UNSET,
                openvpn3.ClientAttentionGroup.USER_PASSWORD,
                openvpn3.ClientAttentionGroup.HTTP_PROXY_CREDS,
                openvpn3.ClientAttentionGroup.PK_PASSPHRASE,
                #openvpn3.ClientAttentionGroup.CHALLENGE_STATIC,
                #openvpn3.ClientAttentionGroup.CHALLENGE_DYNAMIC,
                #openvpn3.ClientAttentionGroup.CHALLENGE_AUTH_PENDING,
            ]
        self.debug(f'Input slot {input_slot.GetLabel()} of type {type}, group {group} is decided {"not " if not result else ""}safe for storage')
        return result
        print(type,group)

    def on_session_event(self, session_id, major, minor, message):
        if session_id not in self.sessions:
            return
        session = self.sessions[session_id]
        major = openvpn3.StatusMajor(major)
        minor = openvpn3.StatusMinor(minor)
        message = str(message)
        self.info(f'Session Event {major} {minor} {message}')
        self.session_statuses[session_id] = {
            'major' : major,
            'minor' : minor,
            'message' : message,
        }
        self.invalid_ui = True

        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CFG_OK == minor:
            try:
                if session_id not in self.sessions_connected:
                    session.Ready()
                    session.Connect()
                    self.sessions_connected.add(session_id)
            except: #TODO: Catch only expected exceptions
                self.debug(traceback.format_exc())
        if openvpn3.StatusMajor.SESSION == major and openvpn3.StatusMinor.SESS_AUTH_URL == minor:
            self.action_auth_url(None, session_id, message)
        if openvpn3.StatusMajor.SESSION == major and openvpn3.StatusMinor.PROC_STOPPED == minor:
            pass
        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CONN_AUTH_FAILED == minor:
            #TODO: Notify authentication failure
            #TODO: Record authentication failure

            self.action_session_disconnect(None, session_id)
            config_id = self.session_configs.get(session_id, None)
            if config_id is not None:
                self.failed_authentications.add(config_id)
                self.action_config_connect(None, config_id)
        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CONN_CONNECTED == minor:
            config_id = self.session_configs.get(session_id, None)
            if config_id is not None and config_id in self.failed_authentications:
                self.failed_authentications.remove(config_id)

        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CONN_FAILED == minor:
            #TODO: Notify connection failure
            #TODO: Record connection failure
            self.action_session_disconnect(None, session_id)
        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CONN_DISCONNECTED == minor:
            pass
        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CONN_DONE == minor:
            pass
        if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CFG_REQUIRE_USER == minor:
            try:
                required_credentials = list()
                for input_slot in session.FetchUserInputSlots():
                    if input_slot.GetTypeGroup()[0] != openvpn3.ClientAttentionType.CREDENTIALS:
                        continue
                    description = str(input_slot.GetLabel())
                    mask = bool(input_slot.GetInputMask())
                    can_store = self.can_store_input_slot(input_slot)
                    required_credentials.append((description, mask, can_store))
                force_ui = False
                config_id = self.session_configs.get(session_id, None)
                if config_id is not None:
                    if config_id in self.failed_authentications:
                        force_ui = True
                self.action_get_credentials(None, session_id, required_credentials, force_ui=force_ui)
            except: #TODO: Catch only expected exceptions
                self.debug(traceback.format_exc())
                #TODO: Catch only expected exceptions
                #TODO: Notify authentication failure
                #TODO: Record authentication failure
                self.action_session_disconnect(None, session_id)
        self.notify_session_change(session_id)

    def action_auth_url(self, _object, session_id, url):
        webbrowser.open_new(url)

    def store_set_credentials(self, config_id, credentials):
        store = self.credential_store[config_id]
        for key, value in credentials.items():
            store[key] = value

    def store_clear_credentials(self, config_id, credentials_keys):
        store = self.credential_store[config_id]
        for key in store.keys():
            if key in credentials_keys:
                del store[key]

    def store_get_credentials(self, config_id):
        credentials = dict()
        store = self.credential_store[config_id]
        for key in store.keys():
            credentials[key] = store[key]
        return credentials

    def action_get_credentials(self, _object, session_id, required_credentials, force_ui=False):
        credentials = dict()
        required_keys = set([ description for description, mask, can_store in required_credentials ])
        config_id = self.session_configs.get(session_id, None)
        if config_id is not None:
            for key, value in self.store_get_credentials(config_id).items():
                if key in required_keys:
                    credentials[key] = value

        require_ui = False
        for key in required_keys:
            if key not in credentials:
                require_ui = True
                break

        if require_ui or force_ui:
            user_inputs = [ CredentialsUserInput(
                    name=description,
                    mask=mask,
                    value=credentials.get(description, None) if can_store else None,
                    can_store=can_store)
                for description, mask, can_store in required_credentials ]


            def on_cancel():
                status = self.session_statuses.get(session_id, None)
                if status is None:
                    return
                major = status['major']
                minor = status['minor']
                if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CFG_REQUIRE_USER == minor:
                    self.action_session_disconnect(None, session_id)
                if session_id in self.session_dialogs:
                    del self.session_dialogs[session_id]

            def on_connect(user_inputs, store):
                credentials = dict([ (ui.name, ui.value) for ui in user_inputs ])
                if store:
                    store_credentials = dict([ (ui.name, ui.value) for ui in user_inputs if ui.can_store ])
                    self.store_set_credentials(config_id, store_credentials)
                else:
                    self.store_clear_credentials(config_id, credentials.keys())
                if session_id in self.session_dialogs:
                    del self.session_dialogs[session_id]
                self.on_session_credentials(session_id, credentials)

            session_name = self.get_session_name(session_id)
            def remain_active():
                self.debug('Remain active')
                if session_id in self.session_statuses:
                    major = self.session_statuses[session_id]['major']
                    minor = self.session_statuses[session_id]['minor']
                    if openvpn3.StatusMajor.CONNECTION == major and openvpn3.StatusMinor.CFG_REQUIRE_USER == minor:
                        return True
                if session_id in self.session_dialogs:
                    del self.session_dialogs[session_id]
                return False
            dialog = construct_credentials_dialog(session_name, user_inputs, on_connect=on_connect, on_cancel=on_cancel, remain_active=remain_active)
            dialog.set_visible(True)
            if session_id in self.session_dialogs:
                self.session_dialogs[session_id].destroy()
            self.session_dialogs[session_id] = dialog
        else:
            self.on_session_credentials(session_id, credentials)

    def on_session_credentials(self, session_id, credentials):
        session = self.sessions[session_id]
        try:
            for input_slot in session.FetchUserInputSlots():
                if input_slot.GetTypeGroup()[0] != openvpn3.ClientAttentionType.CREDENTIALS:
                    continue
                input_slot.ProvideInput(credentials.get(input_slot.GetLabel(), ''))
            session.Ready()
            session.Connect()
            self.sessions_connected.add(session_id)
        except: #TODO: Catch only expected exceptions
            self.debug(traceback.format_exc())
            self.action_session_disconnect(None, session_id)

    def on_schedule(self):
        self.debug(f'Schedule')
        if self.last_invalid + 30 < time.monotonic():
            self.debug('Forced refresh of sessions')
            self.invalid_sessions = True
        if self.invalid_sessions:
            self.last_invalid = time.monotonic()
            self.refresh_sessions()
        if self.invalid_ui:
            self.refresh_ui()
        self.multi_notifier.update()
        if self.startup_config_id or self.startup_config_name:
            config_id = self.startup_config_id or self.name_configs.get(self.startup_config_name, None)
            if config_id and len(self.config_sessions[config_id]) == 0:
                self.debug(f'Starting config {config_id} as requested in startup settings.')
                self.action_config_connect(None, config_id)
            self.startup_config_id = None
            self.startup_config_name = None
        GLib.timeout_add(1000, self.on_schedule)

    def action_config_connect(self, _object, config_id):
        self.info(f'Connect Config {config_id}')
        if config_id not in self.configs:
            return
        try:
            session = self.session_manager.NewTunnel(self.configs[config_id])
            self.settings.set_string('most-recent-configuration-id', config_id)
        except: #TODO: Catch only expected exceptions
            self.debug(traceback.format_exc())
            pass

    def action_config_remove(self, _object, config_id):
        self.info(f'Remove Config {config_id}')
        if config_id not in self.configs:
            return
        try:
            def on_remove():
                if config_id not in self.configs:
                    return
                self.configs[config_id].Remove()
                self.invalid_sessions = True
            dialog = construct_configuration_remove_dialog(name=self.get_config_name(config_id), on_remove=on_remove)
            dialog.set_visible(True)
        except: #TODO: Catch only expected exceptions
            self.debug(traceback.format_exc())
            pass

    def action_session_connect(self, _object, session_id):
        self.info(f'Connect Session {session_id}')
        if session_id not in self.sessions:
            return
        try:
            self.sessions[session_id].Connect()
        except: #TODO: Catch only expected exceptions
            self.debug(traceback.format_exc())
            pass

    def action_session_pause(self, _object, session_id):
        self.info(f'Pause Session {session_id}')
        if session_id not in self.sessions:
            return
        try:
            self.sessions[session_id].Pause()
        except: #TODO: Catch only expected exceptions
            self.debug(traceback.format_exc())
            pass

    def action_session_resume(self, _object, session_id):
        self.info(f'Resume Session {session_id}')
        if session_id not in self.sessions:
            return
        try:
            self.sessions[session_id].Resume()
        except: #TODO: Catch only expected exceptions
            self.debug(traceback.format_exc())
            pass

    def action_session_restart(self, _object, session_id):
        self.info(f'Restart Session {session_id}')
        if session_id not in self.sessions:
            return
        try:
            self.sessions[session_id].Restart()
        except: #TODO: Catch only expected exceptions
            self.debug(traceback.format_exc())
            pass

    def action_session_disconnect(self, _object, session_id):
        self.info(f'Disconnect Session {session_id}')
        if session_id not in self.sessions:
            return
        try:
            self.sessions[session_id].Disconnect()
        except: #TODO: Catch only expected exceptions
            self.debug(traceback.format_exc())
            pass

    def on_config_import(self, name, path):
        self.info(f'Import Config {name} {path}')
        try:
            try:
                config_description = pathlib.Path(path).read_text()
            except FileNotFoundError:
                self.error(
                    msg=f"Configuration file not found: {path}",
                    notify=False,
                    dialog=True,
                    title="Configuration Import Failed"
                )
                return
            except PermissionError:
                self.error(
                    msg=f"Permission denied accessing file: {path}",
                    notify=False,
                    dialog=True,
                    title="Configuration Import Failed"
                )
                return
            except UnicodeDecodeError:
                self.error(
                    msg=f"File encoding error: {path}\nUnable to read file as text. Please check if this is a valid OpenVPN configuration file.",
                    notify=False,
                    dialog=True,
                    title="Configuration Import Failed"
                )
                return
            except OSError as e:
                self.error(
                    msg=f"Error reading file: {path}\n{str(e)}",
                    notify=False,
                    dialog=True,
                    title="Configuration Import Failed"
                )
                return
            try:
                import_args = dict()
                if self.manager_version > 20:
                    # system_tag arrived in openvpn3-linux v21
                    import_args['system_tag'] = APPLICATION_SYSTEM_TAG
                config_obj = self.config_manager.Import(name, config_description, single_use=False, persistent=True, **import_args)
            except dbus.exceptions.DBusException as excp:
                msg = excp.get_dbus_message()
                msg = re.sub(r'^.*GDBus.Error:[^\s]*', '', msg).strip()
                self.error(
                    msg=f"Failed to import configuration {name}:\n{msg}",
                    notify=False,
                    dialog=True,
                    title="Configuration Import Failed"
                )
                return
            except Exception as e:
                self.error(
                    msg=f"Unexpected error during configuration import:\n{str(e)}",
                    notify=False,
                    dialog=True,
                    title="Configuration Import Failed"
                )
                return
            if self.manager_version >= 22:
                try:
                    v = config_obj.Validate()
                except dbus.exceptions.DBusException as excp:
                    msg = excp.get_dbus_message()
                    msg = re.sub(r'^.*GDBus.Error:[^\s]*', '', msg).strip()
                    self.error(
                        msg=f"OpenVPN Config {name} imported from {path} failed validation:\n{msg}",
                        notify=False,
                        dialog=True,
                        title="Configuration Import Failed"
                    )
                    self.info(f'Removing Config {name}')
                    config_obj.Remove()
                    return

            self.invalid_sessions = True
            self.info(msg=f'Successfully imported config {name} from {path}', notify=True)
        except:
            self.debug(traceback.format_exc())
            self.error(
                msg=f"Unexpected error importing configuration {name} from {path}",
                notify=False,
                dialog=True,
                title="Configuration Import Failed"
            )

    def action_config_import(self, _object):
        self.info(f'Import Config')
        dialog = construct_configuration_select_dialog(on_import=self.on_config_import)
        dialog.set_visible(True)

    def action_config_open(self, path):
        self.info(f'Import Config {path}')
        dialog = construct_configuration_import_dialog(path=path, on_import=self.on_config_import)
        dialog.set_visible(True)

    def action_about(self, _object):
        self.info(f'About')
        dialog = construct_about_dialog()
        dialog.set_visible(True)

    def action_quit(self, _object):
        self.info(f'Quit')
        self.release()

    def logging_notify(self, msg, title=f'{APPLICATION_NAME}', icon='active'):
        icon = f'{APPLICATION_NAME}-{icon}'
        self.multi_notifier.new_notifier(
            identifier = 'logging',
            title = title,
            body = msg,
            icon = icon,
            active = True,
            timespan = 2,
            mute_repetitions = False,
        )
        self.multi_notifier.update()

    def debug(self, msg, notify=False, *args, **kwargs):
        logging.debug(msg, *args, **kwargs)
        if notify:
            self.logging_notify(msg)

    def info(self, msg, notify=False, dialog=False, title=None, *args, **kwargs):
        logging.info(msg, *args, **kwargs)
        if notify:
            self.logging_notify(msg)

        if dialog:
            show_info_notification(title=title, message=msg)

    def warning(self, msg, notify=False, dialog=False, title=None, *args, **kwargs):
        logging.warning(msg, *args, **kwargs)
        if notify:
            self.logging_notify(msg, icon='active-error')

        if dialog:
            show_warning_notification(title=title, message=msg)

    def error(self, msg, notify=False, dialog=False, title=None, *args, **kwargs):
        logging.error(msg, *args, **kwargs)
        if notify:
            self.logging_notify(msg, icon="active-error")

        if dialog:
            show_error_dialog(title=title, message=msg)
